// Native
const path = require('path')
const {parse} = require('url')

// Packages
const micro = require('micro')
const auth = require('basic-auth')
const {red} = require('chalk')
const fs = require('fs-promise')
const pathType = require('path-type')
const mime = require('mime-types')
const stream = require('send')

// Ours
const renderDirectory = require('./render')

module.exports = async (req, res, flags, current, ignoredFiles) => {
  if (flags.auth) {
    const credentials = auth(req)

    if (!process.env.SERVE_USER || !process.env.SERVE_PASSWORD) {
      console.error(red('You are running serve with basic auth but did not set the SERVE_USER and SERVE_PASSWORD environment variables.'))
      process.exit(1)
    }

    if (!credentials || credentials.name !== process.env.SERVE_USER || credentials.pass !== process.env.SERVE_PASSWORD) {
      res.statusCode = 401
      res.setHeader('WWW-Authenticate', 'Basic realm="User Visible Realm"')
      return micro.send(res, 401, 'Access Denied')
    }
  }

  const {pathname} = parse(req.url)
  const assetDir = path.normalize(process.env.ASSET_DIR)

  let related = path.parse(path.join(current, pathname))

  if (related.dir.indexOf(assetDir) > -1) {
    const relative = path.relative(assetDir, pathname)
    related = path.parse(path.join(__dirname, '/../assets', relative))
  }

  related = decodeURIComponent(path.format(related))

  const relatedExists = fs.existsSync(related)
  let notFoundResponse = 'Not Found'

  try {
    const custom404Path = path.join(current, '/404.html')
    notFoundResponse = await fs.readFile(custom404Path, 'utf-8')
  } catch (err) {}

  if (!relatedExists && flags.single === undefined) {
    return micro.send(res, 404, notFoundResponse)
  }

  // Check if directory
  if (relatedExists && await pathType.dir(related)) {
    // Normalize path to trailing slash
    // Otherwise problems like #70 will occur
    if (req.url.substr(-1) !== '/') {
      const newPath = req.url + '/'

      res.writeHead(302, {
        Location: newPath
      })

      res.end()
      return
    }

    let indexPath = path.join(related, '/index.html')
    res.setHeader('Content-Type', mime.contentType(path.extname(indexPath)))

    if (!fs.existsSync(indexPath)) {
      // Try to render the current directory's content
      const port = flags.port || req.socket.localPort
      const renderedDir = await renderDirectory(port, current, related, ignoredFiles)

      // If it works, send the directory listing to the user
      if (renderedDir) {
        return micro.send(res, 200, renderedDir)
      }

      // And if it doesn't, see if it's a single page application
      // If that's not true either, send an error
      if (!flags.single) {
        return micro.send(res, 404, notFoundResponse)
      }

      // But IF IT IS true, load the SPA's root index file
      indexPath = path.join(current, '/index.html')
    }

    return stream(req, indexPath).pipe(res)
  }

  if (!fs.existsSync(related) && flags.single) {
    const indexPath = path.join(current, '/index.html')
    return stream(req, indexPath).pipe(res)
  }

  // Serve files without a mime type as text
  stream.mime.default_type = 'text/plain'

  return stream(req, related, {
    dotfiles: 'allow'
  }).pipe(res)
}
